@@ -26,6 +26,7 @@ from typing import (Any, Iterable, List, Mapping, NamedTuple, Optional,
                     Sequence, Tuple, Type, TypeVar, Union)
 from warnings import warn
 
+from hypothesis import assume
 from hypothesis import strategies as st
 from hypothesis.errors import HypothesisWarning, InvalidArgument
 from hypothesis.internal.conjecture import utils as cu
@@ -45,6 +46,7 @@ __all__ = [
     "valid_tuple_axes",
     "broadcastable_shapes",
     "mutually_broadcastable_shapes",
+    "indices",
 ]
 
 
@@ -55,6 +57,7 @@ Float = TypeVar("Float")
 DataType = Union[Boolean, SignedInteger, UnsignedInteger, Float]
 Array = TypeVar("Array")
 Shape = Tuple[int, ...]
+BasicIndex = Tuple[Union[int, slice, None, "ellipsis"], ...]  # noqa: F821
 
 
 class BroadcastableShapes(NamedTuple):
@@ -62,12 +65,11 @@ class BroadcastableShapes(NamedTuple):
     result_shape: Shape
 
 
-DTYPE_NAMES = [
-    "bool",
-    "int8", "int16", "int32", "int64",
-    "uint8", "uint16", "uint32", "uint64",
-    "float32", "float64",
-]
+SIGNED_INT_NAMES = ["int8", "int16", "int32", "int64"]
+UNSIGNED_INT_NAMES = ["uint8", "uint16", "uint32", "uint64"]
+ALL_INT_NAMES = SIGNED_INT_NAMES + UNSIGNED_INT_NAMES
+FLOAT_NAMES = ["float32", "float64"]
+DTYPE_NAMES = ["bool"] + ALL_INT_NAMES + FLOAT_NAMES
 
 
 def partition_attributes_and_stubs(
@@ -91,8 +93,8 @@ def infer_xp_is_compliant(xp):
         array.__array_namespace__()
     except AttributeError:
         warn(
-            f"Could not determine whether module {xp.__name__}"
-            " is an Array API library",
+            f"Could not determine whether module {xp.__name__} "
+            "is an Array API library",
             HypothesisWarning,
         )
 
@@ -113,8 +115,8 @@ def check_xp_attributes(xp, attributes: List[str]):
 def warn_on_missing_dtypes(xp, stubs: List[str]):
     f_stubs = ", ".join(stubs)
     warn(
-        f"Array module {xp.__name__} does not have"
-        f" the following dtypes in its namespace: {f_stubs}.",
+        f"Array module {xp.__name__} does not have "
+        f"the following dtypes in its namespace: {f_stubs}.",
         HypothesisWarning,
     )
 
@@ -125,8 +127,48 @@ def order_check(name, floor, min_, max_):
     if min_ > max_:
         raise InvalidArgument(f"min_{name}={min_} is larger than max_{name}={max_}")
 
-# Note NumPy supports non-array scalars which hypothesis.extra.numpy.from_dtype
-# utilises, but this from_dtype() method returns just base strategies.
+
+def find_dtype_builtin_family(
+        xp, dtype: Type[DataType]
+) -> Tuple[Type[Union[bool, int, float]], List[str]]:
+    builtin_family = None
+    stubs = []
+
+    try:
+        bool_dtype = xp.bool
+        if dtype == bool_dtype:
+            builtin_family = bool
+    except AttributeError:
+        stubs.append("bool")
+
+    int_dtypes, int_stubs = partition_attributes_and_stubs(xp, ALL_INT_NAMES)
+    if dtype in int_dtypes:
+        builtin_family = int
+    stubs.extend(int_stubs)
+
+    float_dtypes, float_stubs = partition_attributes_and_stubs(xp, FLOAT_NAMES)
+    if dtype in float_dtypes:
+        builtin_family = float
+    stubs.extend(float_stubs)
+
+    return builtin_family, stubs
+
+
+def dtype_from_name(xp, name: str) -> Type[DataType]:
+    if name in DTYPE_NAMES:
+        try:
+            return getattr(xp, name)
+        except AttributeError as e:
+            raise InvalidArgument(
+                f"Array module {xp.__name__} does not have "
+                f"dtype {name} in its namespace"
+            ) from e
+    else:
+        f_valid_dtypes = ", ".join(DTYPE_NAMES)
+        raise InvalidArgument(
+            f"{name} is not a valid Array API data type "
+            f"(pick from: {f_valid_dtypes})"
+        )
 
 
 def from_dtype(
@@ -143,32 +185,15 @@ def from_dtype(
     """Creates a strategy which can generate any castable value of the given
     Array API dtype."""
     infer_xp_is_compliant(xp)
+    check_xp_attributes(xp, ["iinfo", "finfo"])
 
     if isinstance(dtype, str):
-        if dtype in DTYPE_NAMES:
-            try:
-                dtype = getattr(xp, dtype)
-                return from_dtype(xp, dtype)
-            except AttributeError as e:
-                raise InvalidArgument(
-                    f"Array module {xp.__name__} does not have"
-                    f" dtype {dtype} in its namespace"
-                ) from e
-        else:
-            f_valid_dtypes = ", ".join(DTYPE_NAMES)
-            raise InvalidArgument(
-                f"{dtype} is not a valid Array API data type"
-                f" (pick from: {f_valid_dtypes})"
-            )
+        dtype = dtype_from_name(xp, dtype)
 
-    stubs = []
+    builtin_family, stubs = find_dtype_builtin_family(xp, dtype)
 
-    try:
-        bool_dtype = xp.bool
-        if dtype == bool_dtype:
-            return st.booleans()
-    except AttributeError:
-        stubs.append("bool")
+    if builtin_family is bool:
+        return st.booleans()
 
     def minmax_values_kw(info):
         kw = {}
@@ -178,8 +203,8 @@ def from_dtype(
         else:
             if min_value < info.min:
                 raise InvalidArgument(
-                    f"dtype {dtype} requires min_value={min_value}"
-                    f" to be at least {info.min}"
+                    f"dtype {dtype} requires min_value={min_value} "
+                    f"to be at least {info.min}"
                 )
             kw["min_value"] = min_value
 
@@ -188,27 +213,19 @@ def from_dtype(
         else:
             if max_value > info.max:
                 raise InvalidArgument(
-                    f"dtype {dtype} requires max_value={max_value}"
-                    f" to be at most {info.max}"
+                    f"dtype {dtype} requires max_value={max_value} "
+                    f"to be at most {info.max}"
                 )
             kw["max_value"] = max_value
 
         return kw
 
-    int_dtypes, int_stubs = partition_attributes_and_stubs(
-        xp, ["int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64"]
-    )
-    if dtype in int_dtypes:
-        check_xp_attributes(xp, ["iinfo"])
+    if builtin_family is int:
         iinfo = xp.iinfo(dtype)
         kw = minmax_values_kw(iinfo)
         return st.integers(**kw)
 
-    float_dtypes, float_stubs = partition_attributes_and_stubs(
-        xp, ["float32", "float64"]
-    )
-    if dtype in float_dtypes:
-        check_xp_attributes(xp, ["finfo"])
+    if builtin_family is float:
         finfo = xp.finfo(dtype)
 
         kw = minmax_values_kw(finfo)
@@ -223,8 +240,6 @@ def from_dtype(
 
         return st.floats(width=finfo.bits, **kw)
 
-    stubs.extend(int_stubs)
-    stubs.extend(float_stubs)
     if len(stubs) > 0:
         warn_on_missing_dtypes(xp, stubs)
 
@@ -241,29 +256,31 @@ class ArrayStrategy(st.SearchStrategy):
         self.unique = unique
         self.array_size = math.prod(shape)
 
-    def check_results_set(self, result, idx, val):
-        if val == val and self.xp.all(result[idx] != val):
-            raise InvalidArgument(
-                f"Generated array element {repr(val)} from"
-                f" strategy {self.elements_strategy}"
-                f" cannot be represented as dtype {self.dtype}."
-                f" Array module {self.xp.__name__} instead"
-                f" represents the element as {result[idx]}.\n"
-                "Consider using a more precise elements strategy,"
-                " for example passing the width argument to floats()."
-            )
+        builtin_family, _ = find_dtype_builtin_family(xp, dtype)
+        self.builtin = builtin_family
 
-    def set_value(self, result, i, val):
+    def set_value(self, result, i, val, strategy=None):
+        strategy = strategy or self.elements_strategy
         try:
             result[i] = val
         except TypeError as e:
-            # TODO check type promotion rules first
             raise InvalidArgument(
-                f"Could not add generated array element {repr(val)}"
+                f"Could not add generated array element {val!r} "
                 f"of dtype {type(val)} to array of dtype {result.dtype}."
             ) from e
+        self.check_set_value(val, result[i], strategy)
 
-        self.check_results_set(result, i, val)
+    def check_set_value(self, val, val_0d, strategy):
+        if ((val == val and val_0d != val) or
+                (self.xp.isfinite(val_0d) and self.builtin(val_0d) != val)):
+            raise InvalidArgument(
+                f"Generated array element {val!r} from strategy {strategy} "
+                f"cannot be represented as dtype {self.dtype}. "
+                f"Array module {self.xp.__name__} instead "
+                f"represents the element as {val_0d!r}. "
+                "Consider using a more precise elements strategy, "
+                "for example passing the width argument to floats()."
+            )
 
     def do_draw(self, data):
         if 0 in self.shape:
@@ -296,17 +313,18 @@ class ArrayStrategy(st.SearchStrategy):
             fill_val = data.draw(self.fill)
             try:
                 result = self.xp.full(self.array_size, fill_val, dtype=self.dtype)
-            except ValueError as e:
+            except Exception as e:
                 raise InvalidArgument(
-                    f"Could not create full array of dtype {self.dtype}"
-                    f" with fill value {repr(fill_val)}"
+                    f"Could not create full array of dtype {self.dtype} "
+                    f"with fill value {fill_val!r}"
                 ) from e
-            self.check_results_set(result, slice(None, None), fill_val)
+            sample = result[0]
+            self.check_set_value(fill_val, sample, strategy=self.fill)
             if self.unique and not self.xp.all(self.xp.isnan(result)):
                 raise InvalidArgument(
-                    f"Array module {self.xp.__name__} did not recognise"
-                    f" fill value as NaN - instead got {repr(result[0])}."
-                    " Cannot fill unique array with non-NaN values."
+                    f"Array module {self.xp.__name__} did not recognise fill "
+                    f"value {fill_val!r} as NaN - instead got {sample!r}. "
+                    "Cannot fill unique array with non-NaN values."
                 )
 
             elements = cu.many(
@@ -353,7 +371,7 @@ def arrays(
     """Returns a strategy for generating Array API arrays."""
 
     infer_xp_is_compliant(xp)
-    check_xp_attributes(xp, ["empty", "full", "all", "isnan", "bool", "reshape"])
+    check_xp_attributes(xp, ["empty", "full", "all", "isnan", "isfinite", "reshape"])
 
     if isinstance(dtype, st.SearchStrategy):
         return dtype.flatmap(
@@ -364,9 +382,11 @@ def arrays(
             lambda s: arrays(xp, dtype, s, elements=elements, fill=fill, unique=unique)
         )
 
+    if isinstance(dtype, str):
+        dtype = dtype_from_name(xp, dtype)
+
     if isinstance(shape, int):
         shape = (shape,)
-
     if not all(isinstance(s, int) for s in shape):
         raise InvalidArgument(
             f"Array shape must be integer in each dimension, provided shape was {shape}"
@@ -419,8 +439,8 @@ def check_dtypes(xp, dtypes: List[Type[DataType]], stubs: List[str]):
     if len(dtypes) == 0:
         f_stubs = ", ".join(stubs)
         raise InvalidArgument(
-            f"Array module {xp.__name__} does not have"
-            f" the following required dtypes in its namespace: {f_stubs}"
+            f"Array module {xp.__name__} does not have "
+            f"the following required dtypes in its namespace: {f_stubs}"
         )
     elif len(stubs) > 0:
         warn_on_missing_dtypes(xp, stubs)
@@ -443,8 +463,8 @@ def boolean_dtypes(xp) -> st.SearchStrategy[Type[Boolean]]:
         return st.just(xp.bool)
     except AttributeError:
         raise InvalidArgument(
-            f"Array module {xp.__name__} does not have"
-            f" a bool dtype in its namespace"
+            f"Array module {xp.__name__} does not have "
+            f"a bool dtype in its namespace"
         ) from None
 
 
@@ -458,8 +478,8 @@ def check_valid_sizes(category: str, sizes: Sequence[int], valid_sizes: Sequence
         f_valid_sizes = ", ".join(str(s) for s in valid_sizes)
         f_invalid_sizes = ", ".join(str(s) for s in invalid_sizes)
         raise InvalidArgument(
-            f"The following sizes are not valid for {category} dtypes:"
-            f" {f_invalid_sizes} (valid sizes: {f_valid_sizes})"
+            f"The following sizes are not valid for {category} dtypes: "
+            f"{f_invalid_sizes} (valid sizes: {f_valid_sizes})"
         )
 
 
@@ -528,7 +548,7 @@ def valid_tuple_axes(
     min_size: int = 0,
     max_size: Optional[int] = None,
 ) -> st.SearchStrategy[Shape]:
-    """Return a strategy for something I don't quite understand yet."""
+    """Return a strategy for axes."""
     if max_size is None:
         max_size = ndim
 
@@ -557,9 +577,6 @@ class MutuallyBroadcastableShapesStrategy(st.SearchStrategy):
         min_side=1,
         max_side=None,
     ):
-        assert 0 <= min_side <= max_side
-        assert 0 <= min_dims <= max_dims
-
         self.base_shape = base_shape
         self.num_shapes = num_shapes
         self.min_dims = min_dims
@@ -646,35 +663,34 @@ def broadcastable_shapes(
     """Return a strategy for generating shapes that are broadcast-compatible
     with the provided shape."""
     check_type(tuple, shape, "shape")
-    strict_check = max_side is None or max_dims is None
     check_type(int, min_side, "min_side")
     check_type(int, min_dims, "min_dims")
 
+    strict_check = max_side is None or max_dims is None
+
     if max_dims is None:
         max_dims = min(32, max(len(shape), min_dims) + 2)
-    else:
-        check_type(int, max_dims, "max_dims")
+    check_type(int, max_dims, "max_dims")
 
     if max_side is None:
-        max_side = max(tuple(shape[-max_dims:]) + (min_side,)) + 2
-    else:
-        check_type(int, max_side, "max_side")
+        max_side = max(shape[-max_dims:] + (min_side,)) + 2
+    check_type(int, max_side, "max_side")
 
     order_check("dims", 0, min_dims, max_dims)
     order_check("side", 0, min_side, max_side)
 
-    if 32 < max_dims:
-        raise InvalidArgument("max_dims cannot exceed 32")
-
-    dims, bound_name = (max_dims, "max_dims") if strict_check else (
-        min_dims, "min_dims")
+    if strict_check:
+        dims = max_dims
+        bound_name = "max_dims"
+    else:
+        dims = min_dims
+        bound_name = "min_dims"
 
     # check for unsatisfiable min_side
     if not all(min_side <= s for s in shape[::-1][:dims] if s != 1):
         raise InvalidArgument(
-            "Given shape=%r, there are no broadcast-compatible "
-            "shapes that satisfy: %s=%s and min_side=%s"
-            % (shape, bound_name, dims, min_side)
+            f"Given shape={shape}, there are no broadcast-compatible "
+            f"shapes that satisfy: {bound_name}={dims} and min_side={min_side}"
         )
 
     # check for unsatisfiable [min_side, max_side]
@@ -682,14 +698,14 @@ def broadcastable_shapes(
         min_side <= 1 <= max_side or all(s <= max_side for s in shape[::-1][:dims])
     ):
         raise InvalidArgument(
-            "Given shape=%r, there are no broadcast-compatible shapes "
-            "that satisfy: %s=%s and [min_side=%s, max_side=%s]"
-            % (shape, bound_name, dims, min_side, max_side)
+            f"Given base_shape={shape}, there are no broadcast-compatible "
+            f"shapes that satisfy all of {bound_name}={dims}, "
+            f"min_side={min_side}, and max_side={max_side}"
         )
 
     if not strict_check:
         # reduce max_dims to exclude unsatisfiable dimensions
-        for n, s in zip(range(max_dims), reversed(shape)):
+        for n, s in zip(range(max_dims), shape[::-1]):
             if s < min_side and s != 1:
                 max_dims = n
                 break
@@ -716,9 +732,8 @@ def mutually_broadcastable_shapes(
     min_side: int = 1,
     max_side: Optional[int] = None,
 ) -> st.SearchStrategy[BroadcastableShapes]:
-    """Return a strategy for generating a specified number of shapes
-    ``num_shapes`` that are mutually-broadcastable with one another and with the
-    provided base shape ``base_shape``."""
+    """Return a strategy for generating a specified number of shapes that are
+    mutually-broadcastable with one another and with the provided base shape."""
 
     check_type(int, num_shapes, "num_shapes")
     if num_shapes < 1:
@@ -751,8 +766,8 @@ def mutually_broadcastable_shapes(
     # check for unsatisfiable min_side
     if not all(min_side <= s for s in base_shape[::-1][:dims] if s != 1):
         raise InvalidArgument(
-            f"Given base_shape={base_shape}, there are no broadcast-compatible"
-            f" shapes that satisfy: {bound_name}={dims} and min_side={min_side}"
+            f"Given base_shape={base_shape}, there are no broadcast-compatible "
+            f"shapes that satisfy: {bound_name}={dims} and min_side={min_side}"
         )
 
     # check for unsatisfiable [min_side, max_side]
@@ -760,9 +775,9 @@ def mutually_broadcastable_shapes(
         min_side <= 1 <= max_side or all(s <= max_side for s in base_shape[::-1][:dims])
     ):
         raise InvalidArgument(
-            f"Given base_shape={base_shape}, there are no broadcast-compatible"
-            f" shapes that satisfy all of {bound_name}={dims},"
-            f" min_side={min_side}, and max_side={max_side}"
+            f"Given base_shape={base_shape}, there are no broadcast-compatible "
+            f"shapes that satisfy all of {bound_name}={dims}, "
+            f"min_side={min_side}, and max_side={max_side}"
         )
 
     if not strict_check:
@@ -785,30 +800,117 @@ def mutually_broadcastable_shapes(
     )
 
 
+class IndexStrategy(st.SearchStrategy):
+    def __init__(self, shape, min_dims, max_dims, allow_ellipsis, allow_none):
+        self.shape = shape
+        self.min_dims = min_dims
+        self.max_dims = max_dims
+        self.allow_ellipsis = allow_ellipsis
+        self.allow_none = allow_none
+
+    def do_draw(self, data):
+        # General plan: determine the actual selection up front with a straightforward
+        # approach that shrinks well, then complicate it by inserting other things.
+        result = []
+        for dim_size in self.shape:
+            if dim_size == 0:
+                result.append(slice(None))
+                continue
+            strategy = st.integers(-dim_size, dim_size - 1) | st.slices(dim_size)
+            result.append(data.draw(strategy))
+        # Insert some number of new size-one dimensions if allowed
+        result_dims = sum(isinstance(idx, slice) for idx in result)
+        while (
+            self.allow_none
+            and result_dims < self.max_dims
+            and (result_dims < self.min_dims or data.draw(st.booleans()))
+        ):
+            i = data.draw(st.integers(0, len(result)))
+            result.insert(i, None)
+            result_dims += 1
+        # Check that we'll have the right number of dimensions; reject if not.
+        # It's easy to do this by construction iff you don't care about shrinking,
+        # which is really important for array shapes.  So we filter instead.
+        assume(self.min_dims <= result_dims <= self.max_dims)
+        # This is a quick-and-dirty way to insert ..., xor shorten the indexer,
+        # but it means we don't have to do any structural analysis.
+        if self.allow_ellipsis and data.draw(st.booleans()):
+            # Choose an index; then replace all adjacent whole-dimension slices.
+            i = j = data.draw(st.integers(0, len(result)))
+            while i > 0 and result[i - 1] == slice(None):
+                i -= 1
+            while j < len(result) and result[j] == slice(None):
+                j += 1
+            result[i:j] = [Ellipsis]
+        else:
+            while result[-1:] == [slice(None, None)] and data.draw(st.integers(0, 7)):
+                result.pop()
+        if len(result) == 1 and data.draw(st.booleans()):
+            # Sometimes generate bare element equivalent to a length-one tuple
+            return result[0]
+        return tuple(result)
+
+
+def indices(
+    shape: Shape,
+    *,
+    min_dims: int = 0,
+    max_dims: Optional[int] = None,
+    allow_ellipsis: bool = True,
+    allow_none: bool = False,
+) -> st.SearchStrategy[BasicIndex]:
+    """Return a strategy for indices."""
+    check_type(tuple, shape, "shape")
+    check_type(bool, allow_ellipsis, "allow_ellipsis")
+    check_type(bool, allow_none, "allow_none")
+    check_type(int, min_dims, "min_dims")
+    if max_dims is None:
+        max_dims = min(max(len(shape), min_dims) + 2, 32)
+    check_type(int, max_dims, "max_dims")
+    order_check("dims", 0, min_dims, max_dims)
+    if not all(isinstance(x, int) and x >= 0 for x in shape):
+        raise InvalidArgument(
+            f"shape={shape!r}, but all dimensions must be of integer size >= 0"
+        )
+    return IndexStrategy(
+        shape,
+        min_dims=min_dims,
+        max_dims=max_dims,
+        allow_ellipsis=allow_ellipsis,
+        allow_none=allow_none,
+    )
+
+
 def get_strategies_namespace(xp) -> SimpleNamespace:
     """Creates a strategies namespace."""
     infer_xp_is_compliant(xp)
 
-    attributes = {
-        "from_dtype": lambda *a, **kw: from_dtype(xp, *a, **kw),
-        "arrays": lambda *a, **kw: arrays(xp, *a, **kw),
-        "array_shapes": array_shapes,
-        "scalar_dtypes": lambda *a, **kw: scalar_dtypes(xp, *a, **kw),
-        "boolean_dtypes": lambda *a, **kw: boolean_dtypes(xp, *a, **kw),
-        "integer_dtypes": lambda *a, **kw: integer_dtypes(xp, *a, **kw),
-        "unsigned_integer_dtypes": lambda *a, **kw: unsigned_integer_dtypes(
-            xp, *a, **kw
+    return SimpleNamespace(
+        from_dtype=update_wrapper(
+            lambda *a, **kw: from_dtype(xp, *a, **kw), from_dtype
         ),
-        "floating_dtypes": lambda *a, **kw: floating_dtypes(xp, *a, **kw),
-        "valid_tuple_axes": valid_tuple_axes,
-        "broadcastable_shapes": broadcastable_shapes,
-        "mutually_broadcastable_shapes": mutually_broadcastable_shapes,
-    }
-
-    globals_ = globals()
-    for name in attributes.keys():
-        orig_func = globals_[name]
-        wrapped_func = attributes[name]
-        attributes[name] = update_wrapper(wrapped_func, orig_func)
-
-    return SimpleNamespace(**attributes)
+        arrays=update_wrapper(
+            lambda *a, **kw: arrays(xp, *a, **kw), arrays
+        ),
+        array_shapes=array_shapes,
+        scalar_dtypes=update_wrapper(
+            lambda *a, **kw: scalar_dtypes(xp, *a, **kw), scalar_dtypes
+        ),
+        boolean_dtypes=update_wrapper(
+            lambda *a, **kw: boolean_dtypes(xp, *a, **kw), boolean_dtypes
+        ),
+        integer_dtypes=update_wrapper(
+            lambda *a, **kw: integer_dtypes(xp, *a, **kw), integer_dtypes
+        ),
+        unsigned_integer_dtypes=update_wrapper(
+            lambda *a, **kw: unsigned_integer_dtypes(xp, *a, **kw),
+            unsigned_integer_dtypes,
+        ),
+        floating_dtypes=update_wrapper(
+            lambda *a, **kw: floating_dtypes(xp, *a, **kw), floating_dtypes
+        ),
+        valid_tuple_axes=valid_tuple_axes,
+        broadcastable_shapes=broadcastable_shapes,
+        mutually_broadcastable_shapes=mutually_broadcastable_shapes,
+        indices=indices,
+    )
